/*----------------------------------------------------------------------------*
 | Headerfile CallbackR1.h                                                    |
 |                                                                            |
 | Version 2.00                                                               |
 |                                                                            |
 |                                                                            |
 | BriDGe::Callback: a tiny library that provides callback functionality.     |
 | Copyright (C) 2002  Bram de Greve                                          |
 |                                                                            |
 | This library is free software; you can redistribute it and/or              |
 | modify it under the terms of the GNU Lesser General Public                 |
 | License as published by the Free Software Foundation; either               |
 | version 2.1 of the License, or (at your option) any later version.         |
 |                                                                            |
 | This library is distributed in the hope that it will be useful,            |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of             |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          |
 | Lesser General Public License for more details.                            |
 |                                                                            |
 | You should have received a copy of the GNU Lesser General Public           |
 | License along with this library; if not, write to the Free Software        |
 | Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  |
 |                                                                            |
 |                                                                            |
 | Contact by e-mail: Bram de Greve <bdegreve@mail.be>                        |
 |                                                                            |
 |                                                                            |
 | History:                                                                   |
 | May 28th, 2002: creation of CallbackR1.h                                   |
 | June 2th, 2002: Getting rid of interal compiler error.                     |
 |                 Yiha!  We've bypassed the system!  mwuahaha!  hihi :)      |
 | June 7th, 2002: Splitting CallbackR1 to multiple files.                    |
 | June 8th, 2002: Revamping the code.                                        |
 |                 Putting stuff under GNU LGPL.                              |
 | June 9th, 2002: Getting rid of boost::shared_ptr so we look less silly,    |
 |                 Getting rid of makeCallback since we use templated         |
 |                 constructors now.                                          |
 |                 Making the call of an empty callback a little more safe    |
 |                 by throwing an exception CallbackEmptyCall.                |
 |                 Getting rid of operator bool and renaming isEmpty to empty.|
 |                 And some other stuff too.  Thank you Dirk Gerrits!         |
 |                 Soit, we have now officially reached version 2.00!         |
 |                                                                            |
 | I want to thank Altair, Jaap Suter & especially Dirk Gerrits (Who has      |
 | spent an entire afternoon on this to help me.  It was a waste of time for  |
 | both of us, but we had a lot of fun, didn't we? :)                         |
 *----------------------------------------------------------------------------*/

// HOW TO USE THIS:
//
// 1. You create a CallbackR0 by using a ... constructor.  Surprise, surprise!
//    It used to be done by maker functions, but we have now templated
//    constructors.  Very cool:
//          int foo(int a);
//          CallbackR1<int, int> callback1(foo);
//
//          class Bar
//          {
//          public:
//              int fun(int a);
//          }
//          Bar bar;
//          CallbackR0<int, int> callback2(bar, Bar::fun);
// 
// 2. You can copy callback objects as much as you like.
//          CallbackR0<int, int> callback3(callback1);
//
//          CallbackR0<int, int> callback4;
//          callback4 = callback2;
//
// 3. To excecute the callback, you evoke the function operator
//          int r = callback1(999);
//
// Note: Although you can allocate callbacks without specifying the exact
//       callback function, you must make sure the callback can do something
//       when you excecute it.  So, you must feed it with appropriate beef 
//       (functions, object/method pair).  If you don't, funny things will
//       happen.  A crash for example.  Luckely for you, a special exception
//       BDG::CallbackEmptyCall will be thrown so that you can catch it.
//       On the other hand, you could also have checked this before the call
//       with the method empty().
//          try
//          {
//              int r = callback1(999);
//          }
//          catch (CallbackEmptyCall)
//          {
//              ...
//          }
//
//          if (!callback1.empty(999))
//          {
//              int r = callback1();
//          }
//
// Note: If you use a object/method pair as callback, you must make sure the 
//       object is alive when calling the callback.  Otherwise the callback
//       will want to do the method on dead beef, and you know funny things
//       happen when it does that.  And you know what that means, don't you?



#ifndef BDG_GUARDIAN_OF_INCLUSION_CALLBACKR1_H
#define BDG_GUARDIAN_OF_INCLUSION_CALLBACKR1_H



// --- OLD INTERFACES ----------------------------------------------------------

#include "CallbackR1Detail.h"
#include "CallbackExceptions.h"
#include <algorithm>                    // std::swap



// --- NEW INTERFACES ----------------------------------------------------------

namespace BDG
{


// THE CALLBACK:
// This is the actuall class you use to hold callbacks:

template <typename R, typename P1>
class CallbackR1
{
public:

    // CONSTRUCTORS & DESTRUCTORS
    
    CallbackR1(): 
        m_dispatcher(0) 
    {
    }

    template <typename FR, typename FP1>
    explicit CallbackR1(FR (*a_function)(FP1))
    {
        if (!a_function) throw CallbackNullArgument(
            "BDG::CallbackR1: tried to construct from null function pointer.");
        using namespace CallbackR1_Detail;
        m_dispatcher = new DispFunction<R, P1, FR, FP1>(a_function);
    }
    
    template<class Object, typename MR, typename MP1>
    CallbackR1(Object* a_object, MR (Object::*a_method)(MP1))
    {
        if (!(a_object && a_method)) throw CallbackNullArgument(
            "BDG::CallbackR1: tried to construct from null object or method "
            "pointer.");
        using namespace CallbackR1_Detail;
        m_dispatcher = new DispMethod<Object, R, P1, MR, MP1>(a_object,
                                                              a_method);
    }

    template<class Object, typename MR, typename MP1>
    CallbackR1(Object* a_object, MR (Object::*a_method)(MP1) const)
    {
        if (!(a_object && a_method)) throw CallbackNullArgument(
            "BDG::CallbackR1: tried to construct from null object or method "
            "pointer.");
        using namespace CallbackR1_Detail;
        m_dispatcher = new DispMethodConst<Object, R, P1, MR, MP1>(a_object, 
                                                                   a_method);
    }

    CallbackR1(const CallbackR1<R, P1>& a_other):
        m_dispatcher(a_other.empty() ? 0 : a_other.m_dispatcher->clone())
    {
    }
    
    
    // OPERATORS

    CallbackR1 operator=(const CallbackR1<R, P1>& a_other)
    {
        CallbackR1<R, P1> temp(a_other);
        swap(temp);
        return *this;
    }           
    
    R operator()(P1 a_p1) const 
    {
        if (empty()) throw CallbackEmptyCall(
            "BDG::CallbackR1: tried to call an empty callback.");
        return (*m_dispatcher)(a_p1); 
    }


    // METHODS

    bool empty() const
    {
        return m_dispatcher == 0;
    }

    void swap(CallbackR1<R, P1>& a_other)
    {
        std::swap(m_dispatcher, a_other.m_dispatcher);
    }

private:

    CallbackR1_Detail::DispBase<R, P1>* m_dispatcher;
};



} // namespace BDG

#endif // Guardian of Inclusion

// --- END OF FILE CallbackR1.h ------------------------------------------------
